﻿# Author: Anthony Stahler
# Last updated: Feb 2, 2023
# License: Free for personal use, for commercial use please reach out to the author via the BestEdit contact page: http://bestedit.net/contact/
# The most recent version of this script can be found at http://bestedit.net/encode-mp4/
#
# I couldn't find a way internally in ffmpeg to reliably change an -ss shift once it had been set by a full audio/video encode.  
# Re-encoding the audio alone didn't seem to work, and I was getting oddly inconsistent results from atrim/adelay and itsoffset depending on the type of original 
# source file at the start of the workflow. So this script uses ffmpeg to extract the ac3 audio, process it with an external program called eac3to, and then rejoin it with
# the original video using ffmpeg, to create a reliably shifted audio.
#
# Please note that this does not overwrite final files, so if you run it twice with one new mp4 in the source directory, it will only process that one file and
# not waste time re-creating the others it has already completed.  So if you want to re-do a file, delete the final mp4 it created from your last run.
# 
# If you're trying to create the same file shifted multiple ways you could save yourself by commenting out these two lines in the script below:
# Remove-Item -LiteralPath ("$origpath$origname1$ac3ext") -Force -Recurse
# Remove-Item -LiteralPath ("$interpath$origname1$ac3ext$w64ext") -Force -Recurse
#
# Those lines are cleaning up early stages of the file extraction and encoding process, and they're identical output files no matter how you're going to shift your final mp4.
# If the script sees those files already present it won't redo that stage, skipping it to save you time.  Just be aware that the w64 files are quite large (10gb ballpark) so 
# make sure you have a very large drive if you choose not to clean them up after each converted file!
#


# --------------------------
# CONSTANTS (don't touch)
# --------------------------

$msperframe = 5.33333
$srctype = "*.mp4"

$ac3ext = ".ac3"
$w64ext = "-intermediate-full.w64"
$mp4ext = ".mp4"

# --------------------------
# VARIABLES 
# --------------------------

$eac3toexe = "C:\eac3to_3.47\eac3to.exe"

# folder with our mp4 files
$srcpath = "C:\test-eac-shift\"

# folder with lots of space for transcoding and re-encoding
$workingpath = "C:\eac3to-shifted\"

# extra milliseconds error added by -ss command in ffmpeg
$extrams = -5.333333
$basems = 0

# --------------------------
# --------------------------

# totalshift is only accepted in round numbers, eac3to doesn't like decimal milliseconds
[int]$totalshift = [math]::Round($basems + $extrams)
$totalshiftconcat = "-" + $totalshift + "ms"
$totalshiftconcatn = $totalshiftconcat #$totalshiftconcatn is for the file name

# this lettering may seem backward because we're matching it with SS for consistency. Negative SS adds more blank frames, similar to apad so it moves audio right in relation to video track
if ($totalshift -lt 0)
{
    # E shift, moving left
    $totalshiftconcatn = "-E" + $totalshift + "ms"
}
elseif ($totalshift -gt 0)
{
    # N shift, moving right
    $totalshiftconcatn = "-N" + $totalshift + "ms"
}

$origpath =  $workingpath + "original-ac3\"
$interpath = $workingpath + "intermediate-w64\"
$shiftpath = $workingpath + "shifted-ac3\"
$finpath = $workingpath + "finalmp4\"

# create the working directories if they don't exist
$silenceoutput = New-Item -ItemType Directory -Force -Path $origpath
$silenceoutput = New-Item -ItemType Directory -Force -Path $interpath
$silenceoutput = New-Item -ItemType Directory -Force -Path $shiftpath
$silenceoutput = New-Item -ItemType Directory -Force -Path $finpath

Get-ChildItem -path ( $srcpath + $srctype ) | foreach {

    $origname1 = $_.Name
    $origname1NOext = $_.BaseName

    $currentDisplayTime = Get-Date

    Write-Output "Processing: '$origname1' at $currentDisplayTime"

    if (![System.IO.File]::Exists("$finpath$origname1$ac3ext$w64ext$totalshiftconcatn$ac3ext$mp4ext"))
    {
        # first we want to extract the ac3 file 
        # the -y switch is to stop ffmpeg from prompting us for info, for example on overwrite of existing files
        # ffmpeg also has an annoying way of displaying output so we need to append 2>$null to silence it
        #for %i in (*.mp4) do ffmpeg -i "%i" -c:a copy "S:\eac3to-shifted\original-ac3\%i.ac3"
        if (![System.IO.File]::Exists("$origpath$origname1$ac3ext"))
        {
            $silenceoutput = &"ffmpeg" "-y" "-i" "`"$srcpath$origname1`"" "-c:a" "copy" "`"$origpath$origname1$ac3ext`"" 2>$null
        }

        # then we want to turn it into a 64 bit wav, because ac3 only allows us to shift per frame which is 6*256 samples (32ms), turning it into a large wav and then back gives us the ability to shift per-ms granularity
        # we could just keep the new extension for each step, but I like the traceback chain it gives us for later to remember what steps a file took if it gets played with multiple times
        #for %i in (*.ac3) do "R:\rshare\eac3to_3.47\eac3to.exe" "%i" "S:\eac3to-shifted\intermediate-w64\%i-intermediate-full.w64" -full
        if (![System.IO.File]::Exists("$interpath$origname1$ac3ext$w64ext"))
        {
            $silenceoutput = &"$eac3toexe" "`"$origpath$origname1$ac3ext`"" "`"$interpath$origname1$ac3ext$w64ext`"" "-full" 
        }

        # now shift by specified ms and then encode back to ac3
        # for %i in (*.w64) do "R:\rshare\eac3to_3.47\eac3to.exe" "%i" "output\%i-N16ms.ac3" -640 -16ms
        if (![System.IO.File]::Exists("$shiftpath$origname1$ac3ext$w64ext$totalshiftconcatn$ac3ext"))
        {
            if ($totalshift -ne 0)
            {
                $silenceoutput = &"$eac3toexe" "`"$interpath$origname1$ac3ext$w64ext`"" "`"$shiftpath$origname1$ac3ext$w64ext$totalshiftconcatn$ac3ext`"" "-640" "$totalshiftconcat"
            }
            else
            {
                $silenceoutput = &"$eac3toexe" "`"$interpath$origname1$ac3ext$w64ext`"" "`"$shiftpath$origname1$ac3ext$w64ext$totalshiftconcatn$ac3ext`"" "-640" 
            }
        }

        # finally combine with original mp4 again
        # for %i in (*.mp4) do ffmpeg -i "S:\eac3to-shifted\output\%i.ac3-intermediate-full.w64-N15ms.ac3" -i "%i" -c:a copy -c:v copy -map 0:a -map 1:v "S:\eac3to-shifted\finalmp4\%i-AC3-N15ms-FULL.mp4"
        $silenceoutput = &"ffmpeg" "-y" "-i" "`"$shiftpath$origname1$ac3ext$w64ext$totalshiftconcatn$ac3ext`"" "-i" "`"$srcpath$origname1`"" "-c:a" "copy" "-c:v" "copy" "-map" "0:a" "-map" "1:v" "`"$finpath$origname1$ac3ext$w64ext$totalshiftconcatn$ac3ext$mp4ext`"" 2>$null

        # now remove all the temporary files and start the process again
        Remove-Item -LiteralPath ("$origpath$origname1$ac3ext") -Force -Recurse
        Remove-Item -LiteralPath ("$interpath$origname1$ac3ext$w64ext") -Force -Recurse
        Remove-Item -LiteralPath ("$shiftpath$origname1$ac3ext$w64ext$totalshiftconcatn$ac3ext") -Force -Recurse
    }

}